"""
This file holds the generic errors which are sensible for several
areas of Ensemble.
"""


class EnsembleError(Exception):
    """All errors in Ensemble are subclasses of this.

    This error should not be raised by itself, though, since it means
    pretty much nothing.  It's useful mostly as something to catch instead.
    """


class IncompatibleVersion(EnsembleError):
    """Raised when there is a mismatch in versions using the topology.

    This mismatch will occur when the /topology node has the key
    version set to a version different from
    ensemble.state.topology.VERSION in the code itself. This scenario
    can occur when a new client accesses an environment deployed with
    previous code, or upon the update of the code in the environment
    itself.

    Although this checking is done at the level of the topology, upon
    every read, the error is defined here because of its
    generality. Doing the check in the topology is just because of the
    centrality of that piece within Ensemble.
    """

    def __init__(self, current, wanted):
        self.current = current
        self.wanted = wanted

    def __str__(self):
        return (
            "Incompatible ensemble protocol versions (found %r, want %r)" % (
                self.current, self.wanted))


class FileNotFound(EnsembleError):
    """Raised when a file is not found, obviously! :-)

    @ivar path: Path of the directory or file which wasn't found.
    """

    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "File was not found: %s" % (self.path,)


class FormulaError(EnsembleError):
    """An error occurred while processing a formula."""

    def __init__(self, path, message):
        self.path = path
        self.message = message

    def __str__(self):
        return "Error processing '%s': %s." % (self.path, self.message)


class FormulaInvocationError(EnsembleError):
    """A formula's hook invocation exited with an error"""

    def __init__(self, path, exit_code):
        self.path = path
        self.exit_code = exit_code

    def __str__(self):
        return "Error processing '%s': exit code %s." % (self.path,
                                                         self.exit_code)


class FileAlreadyExists(EnsembleError):
    """Raised when something refuses to overwrite an existing file.

    @ivar path: Path of the directory or file which wasn't found.
    """

    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "File already exists, won't overwrite: %s" % (self.path,)


class InvalidEnsembleHeaderValue(EnsembleError):
    """Raised when the ensemble: option value in the YAML file is unexpected.

    @ivar expected: The value which was actually expected.
    @ivar found: The value which was found.
    """

    def __init__(self, path, expected, found):
        self.expected = expected
        self.found = found
        self.path = path

    def __str__(self):
        return ("Expected a YAML file with an 'ensemble: %s' "
                "header, found 'ensemble: %s': %s"
                % (self.expected, self.found, self.path))


class NoConnection(EnsembleError):
    """Raised when the CLI is unable to establish a Zookeeper connection."""


class InvalidHost(NoConnection):
    """Raised when the CLI cannot connect to ZK because of an invalid host."""


class InvalidUser(NoConnection):
    """Raised when the CLI cannot connect to ZK because of an invalid user."""


class EnvironmentNotFound(NoConnection):
    """Raised when the ensemble environment cannot be found."""

    def __init__(self, info="no details available"):
        self._info = info

    def __str__(self):
        return "Ensemble environment not found: %s" % self._info


class EnvironmentPending(NoConnection):
    """Raised when the ensemble environment is not accessible."""

    def __init__(self, info="no details available"):
        self._info = info

    def __str__(self):
        return "Ensemble environment is not accessible: %s" % self._info


class ProviderError(EnsembleError):
    """Raised when an exception occurs in a provider."""


class MachinesNotFound(ProviderError):
    """Raised when a provider can't fulfil a request for machines."""

    def __init__(self, instance_ids):
        self.instance_ids = list(instance_ids)

    def __str__(self):
        return "Cannot find machine%s: %s" % (
            "" if len(self.instance_ids) == 1 else "s",
            ", ".join(self.instance_ids))


class ProviderInteractionError(ProviderError):

    def __init__(self, error):
        self.error = error

    def __str__(self):
        return "ProviderError: Interaction with machine provider failed: %r" \
               % self.error


class CannotTerminateMachine(EnsembleError):
    """Cannot terminate machine because of some reason"""

    def __init__(self, id, reason):
        self.id = id
        self.reason = reason

    def __str__(self):
        return "Cannot terminate machine %d: %s" % (self.id, self.reason)
