import copy
from operator import itemgetter

from ensemble.environment.errors import EnvironmentsConfigError
from ensemble.providers.common.bootstrap import Bootstrap
from ensemble.providers.common.findzookeepers import find_zookeepers
from ensemble.providers.common.state import SaveState, LoadState
from ensemble.providers.common.utils import get_user_authorized_keys


class MachineProviderBase(object):
    """Base class supplying common functionality for MachineProviders.

    To write a working subclass, you will need to override the following
    methods:

        * connect
        * get_file_storage
        * start_machine
        * get_machines
        * shutdown_machines
        * open_port
        * close_port
        * get_opened_ports

    You may want to override the following methods, but you should be careful
    to call MachineProviderBase's implementation (or be very sure you don't
    need to:

        * __init__
        * get_serialization_data

    You probably shouldn't override anything else.
    """

    def __init__(self, environment_name, config):
        if ("authorized-keys-path" in config and
            "authorized-keys" in config):
            raise EnvironmentsConfigError(
                "Environment config cannot define both authorized-keys "
                "and authorized-keys-path. Pick one!")

        self.environment_name = environment_name
        self.config = config

    def get_serialization_data(self):
        """Get provider configuration suitable for serialization."""
        data = copy.deepcopy(self.config)
        data["authorized-keys"] = get_user_authorized_keys(data)
        # Not relevant, on a remote system.
        data.pop("authorized-keys-path", None)
        return data

    #================================================================
    # Subclasses need to implement their own versions of everything
    # in the following block

    def connect(self, share=False):
        """Connect to the zookeeper ensemble running in the machine provider.

        @param share: Requests sharing of the connection with other clients
            attempting to connect to the same provider, if that's feasible.

        returns an open C{txzookeeper.client.ZookeeperClient} and a
        C{ensemble.storage.connection.TunnelProtocol}
        """
        raise NotImplementedError()

    def get_file_storage(self):
        """Retrieve the provider C{FileStorage} abstraction."""
        raise NotImplementedError()

    def start_machine(self, machine_data, master=False):
        """Start a machine in the provider.

        @param machine_data: a dictionary of data to pass along to the newly
           launched machine.

        @param master: if True, machine will initialize the ensemble admin
            and run a provisioning agent.
        """
        raise NotImplementedError()

    def get_machines(self, instance_ids=()):
        """List machines running in the provider.

        @param instance_ids: ids of instances you want to get. Leave empty
            to list all machines owned by this provider.

        @return: a list of provider-specific ProviderMachines.

        @raise: MachinesNotFound
        """
        raise NotImplementedError()

    def shutdown_machines(self, requested_machines=()):
        """Terminate machines associated with this provider.

        @param requested_machines: list of machines to shut down; leave
            empty to shut down all associated machines.
        """
        raise NotImplementedError()

    def open_port(self, machine, machine_id, port, protocol="tcp"):
        """Authorizes `port` using `protocol` for `machine`."""
        raise NotImplementedError()

    def close_port(self, machine, machine_id, port, protocol="tcp"):
        """Revokes `port` using `protocol` for `machine`."""
        raise NotImplementedError()

    def get_opened_ports(self, machine, machine_id):
        """Returns a set of open (port, protocol) pairs for `machine`."""
        raise NotImplementedError()

    #================================================================
    # Subclasses will not generally need to override the methods in
    # this block

    def get_zookeeper_machines(self):
        """Find running zookeeper instances.

        @return: the first valid instance found as a single element list.

        @raise: EnvironmentNotFound
        """
        return find_zookeepers(self)

    def bootstrap(self):
        """Bootstrap an ensemble server in the provider."""
        return Bootstrap(self).run()

    def get_machine(self, instance_id):
        """Retrieve a provider machine by instance id.

        @param instance_id: instance_id of the Machine you want to get.

        @raise: MachinesNotFound
        """
        d = self.get_machines([instance_id])
        d.addCallback(itemgetter(0))
        return d

    def shutdown_machine(self, machine):
        """Terminate one machine associated with this provider.

        @param machine: machine to shut down.

        @raise: MachinesNotFound
        """
        d = self.shutdown_machines([machine])
        d.addCallback(itemgetter(0))
        return d

    def destroy_environment(self):
        """Clear ensemble state and terminate all associated machines"""
        return self.shutdown_machines()

    def save_state(self, state):
        """Save state to the provider.

        @param state
        @type dict
        """
        return SaveState(self).run(state)

    def load_state(self):
        """Load state from the provider.

        @return: a dictionary.
        """
        return LoadState(self).run()
