import logging
import tempfile

from twisted.internet.defer import fail, inlineCallbacks, succeed

from ensemble.errors import EnvironmentNotFound, ProviderError
from ensemble.lib.testing import TestCase
from ensemble.providers.common.base import MachineProviderBase
from ensemble.providers.common.launch import (
    BOOTSTRAP_PACKAGES, DEFAULT_PACKAGES, DEFAULT_REPOSITORIES, LaunchMachine)
from ensemble.providers.dummy import DummyMachine, FileStorage

DEFAULT_INSTALL = ["sudo apt-get -y install ensemble"]
CREATE_DIRS = ["sudo mkdir -p /var/lib/ensemble",
               "sudo mkdir -p /var/log/ensemble"]


class DummyLaunchMachine(LaunchMachine):

    def start_machine(self, variables, data):
        if self._master:
            name = "bootstrapped-instance-id"
        else:
            name = "some-instance-id"
        return [DummyMachine(name)]

    def get_instance_id_command(self):
        return "$(magic)"


class WorkingFileStorage(FileStorage):

    def __init__(self):
        super(WorkingFileStorage, self).__init__(tempfile.mkdtemp())


class UnwritableFileStorage(object):

    def put(self, name, f):
        return fail(Exception("oh noes"))


def get_provider(launch_class=DummyLaunchMachine, config=None, zookeeper=True,
                 file_storage_class=WorkingFileStorage):

    class DummyProvider(MachineProviderBase):

        def __init__(self, config):
            super(DummyProvider, self).__init__("venus", config)
            self._file_storage = file_storage_class()

        def get_zookeeper_machines(self):
            # this is mocked out to avoid insane complexity
            if isinstance(zookeeper, Exception):
                return fail(zookeeper)
            return succeed(
               [DummyMachine("zookeeper-instance-id",
                             private_dns_name="zookeeper.internal")])

        def get_file_storage(self):
            return self._file_storage

        def start_machine(self, machine_data, master=False):
            return launch_class(self, master).run(machine_data)

    create_config = {"authorized-keys": "abc"}
    if config is not None:
        create_config.update(config)
    return DummyProvider(create_config)


def get_launch(launch_class=DummyLaunchMachine, config=None,
               zookeeper=True, master=False):
    provider = get_provider(launch_class, config, zookeeper)
    return launch_class(provider, master=master)


class LaunchMachineTest(TestCase):

    def _verify_machines(self, machines):
        (machine,) = machines
        self.assertTrue(isinstance(machine, DummyMachine))
        self.assertEquals(machine.instance_id, "some-instance-id")

    def test_no_machine_id(self):
        launch = get_launch()
        d = launch.run({})
        self.assertFailure(d, ProviderError)
        return d

    def test_basic_run(self):
        launch = get_launch()
        d = launch.run({"machine-id": "machine-77"})
        d.addCallback(self._verify_machines)
        return d

    @inlineCallbacks
    def test_bootstrap_run(self):
        provider = get_provider(config={"admin-secret": "whatever"})
        launch = DummyLaunchMachine(provider, master=True)
        yield launch.run({"machine-id": "machine-32767"})
        saved_state = yield provider.load_state()
        self.assertEquals(
            saved_state, {"zookeeper-instances": ["bootstrapped-instance-id"]})

    def test_get_machine_variables_normal(self):
        launch = get_launch()
        d = launch.get_machine_variables({"machine-id": "machine-999"})

        def verify_vars(vars):
            expect_vars = {
                "authorized_keys": "abc",
                "ensemble-zookeeper-hosts": "zookeeper.internal:2181",
                "machine-id": "machine-999",
                "packages": DEFAULT_PACKAGES,
                "repositories": DEFAULT_REPOSITORIES,
                "scripts": DEFAULT_INSTALL + CREATE_DIRS +
                           ["ENSEMBLE_MACHINE_ID=machine-999 "
                            "ENSEMBLE_ZOOKEEPER=zookeeper.internal:2181 "
                            "python -m ensemble.agents.machine -n "
                            "--logfile=/var/log/ensemble/machine-agent.log "
                            "--pidfile=/var/run/ensemble/machine-agent.pid"]}
            self.assertEquals(vars, expect_vars)
        d.addCallback(verify_vars)
        return d

    def test_get_machine_variables_with_ensemble_branch(self):
        branch_name = "lp:~my/ensemble/branch"
        launch = get_launch(config={"ensemble-branch": branch_name})
        d = launch.get_machine_variables({"machine-id": "machine-303"})

        def verify_vars(vars):
            scripts = vars["scripts"]
            install_scripts = ["sudo apt-get install -y python-txzookeeper",
                               "sudo mkdir -p /usr/lib/ensemble",
                               "cd /usr/lib/ensemble && "
                               "sudo /usr/bin/bzr co "
                               + branch_name + " ensemble",
                               "cd /usr/lib/ensemble/ensemble && "
                               "sudo python setup.py develop"]
            install_count = len(install_scripts)
            self.assertEquals(scripts[:install_count], install_scripts)

            other_scripts = scripts[install_count:]
            self.assertEquals(other_scripts,
                              CREATE_DIRS + [
                              "ENSEMBLE_MACHINE_ID=machine-303 "
                              "ENSEMBLE_ZOOKEEPER=zookeeper.internal:2181 "
                              "python -m ensemble.agents.machine -n "
                              "--logfile=/var/log/ensemble/machine-agent.log "
                              "--pidfile=/var/run/ensemble/machine-agent.pid"])
        d.addCallback(verify_vars)
        return d

    def test_get_machine_variables_bootstrap(self):
        launch = get_launch(config={"admin-secret": "SEEKRIT"}, master=True)
        d = launch.get_machine_variables({"machine-id": "machine-757"})

        def verify_vars(vars):
            expect_vars = {
                "authorized_keys": "abc",
                "ensemble-zookeeper-hosts": "localhost:2181",
                "machine-id": "machine-757",
                "packages": DEFAULT_PACKAGES + BOOTSTRAP_PACKAGES,
                "repositories": DEFAULT_REPOSITORIES}
            for key, value in expect_vars.items():
                self.assertEquals(vars[key], value)

            scripts = vars["scripts"]
            self.assertEquals(scripts[:3], DEFAULT_INSTALL + CREATE_DIRS)
            admin_init, machine_agent, provision_agent = scripts[3:]

            self.assertTrue(admin_init.startswith(
                "ensemble-admin initialize "
                '--instance-id=$(magic) --admin-identity="admin:'))
            self.assertFalse("SEEKRIT" in admin_init)

            expect_machine = (
                "ENSEMBLE_MACHINE_ID=machine-757 "
                "ENSEMBLE_ZOOKEEPER=localhost:2181 "
                "python -m ensemble.agents.machine -n "
                "--logfile=/var/log/ensemble/machine-agent.log "
                "--pidfile=/var/run/ensemble/machine-agent.pid")
            self.assertEquals(machine_agent, expect_machine)

            expect_provision = (
                "ENSEMBLE_ZOOKEEPER=localhost:2181 "
                "python -m ensemble.agents.provision -n "
                "--logfile=/var/log/ensemble/provision-agent.log "
                "--pidfile=/var/run/ensemble/provision-agent.pid")
            self.assertEquals(provision_agent, expect_provision)
        d.addCallback(verify_vars)
        return d


class BootstrapTest(TestCase):

    def test_bootstrap_unknown_error(self):
        class BBQError(Exception):
            pass
        provider = get_provider(zookeeper=BBQError())
        d = provider.bootstrap()
        self.assertFailure(d, BBQError)
        return d

    def test_bootstrap_unwritable_storage(self):
        provider = get_provider(zookeeper=EnvironmentNotFound(),
                                file_storage_class=UnwritableFileStorage)
        d = provider.bootstrap()
        self.assertFailure(d, ProviderError)
        return d

    def test_bootstrap_no_launch(self):
        log = self.capture_logging("ensemble.common", level=logging.DEBUG)
        provider = get_provider()
        d = provider.bootstrap()

        def verify_machines(machines):
            (machine,) = machines
            self.assertTrue(isinstance(machine, DummyMachine))
            self.assertEquals(machine.instance_id, "zookeeper-instance-id")
            log_text = log.getvalue()
            self.assertIn(
                "Ensemble environment previously bootstrapped", log_text)
            self.assertNotIn("Launching", log_text)
        d.addCallback(verify_machines)
        return d

    def test_bootstrap_launch(self):
        log = self.capture_logging("ensemble.common", level=logging.DEBUG)
        provider = get_provider(config={"admin-secret": "SEEKRIT"},
                                zookeeper=EnvironmentNotFound())
        d = provider.bootstrap()

        def verify_machines(machines):
            (machine,) = machines
            self.assertTrue(isinstance(machine, DummyMachine))
            self.assertEquals(machine.instance_id, "bootstrapped-instance-id")
            log_text = log.getvalue()
            self.assertIn("Launching Ensemble bootstrap instance", log_text)
            self.assertNotIn("previously bootstrapped", log_text)
        d.addCallback(verify_machines)
        return d
