import os
import re

from twisted.internet.defer import inlineCallbacks, returnValue
from txaws.ec2.exception import EC2Error
from txaws.service import AWSServiceRegion

from ensemble.errors import (
    MachinesNotFound, ProviderError, ProviderInteractionError)
from ensemble.providers.common.base import MachineProviderBase
from ensemble.providers.common.utils import convert_unknown_error

from .connect import EC2Connect
from .files import FileStorage
from .launch import EC2LaunchMachine
from .machine import EC2ProviderMachine, machine_from_instance
from .securitygroup import (
    open_provider_port, close_provider_port, get_provider_opened_ports,
    remove_security_groups)
from .utils import get_region_uri, log


class MachineProvider(MachineProviderBase):

    def __init__(self, environment_name, config):
        super(MachineProvider, self).__init__(environment_name, config)

        if not config.get("ec2-uri"):
            ec2_uri = get_region_uri(config.get("region", "us-east-1"))
        else:
            ec2_uri = config.get("ec2-uri")

        self._service = AWSServiceRegion(
            access_key=config.get("access-key", ""),
            secret_key=config.get("secret-key", ""),
            ec2_uri=ec2_uri,
            s3_uri=config.get("s3-uri", ""))
        self.s3 = self._service.get_s3_client()
        self.ec2 = self._service.get_ec2_client()

    def get_serialization_data(self):
        """Get provider configuration suitable for serialization.

        Also extracts credential information from the environment.
        """
        data = super(MachineProvider, self).get_serialization_data()
        data.setdefault("access-key", os.environ.get("AWS_ACCESS_KEY_ID"))
        data.setdefault("secret-key", os.environ.get("AWS_SECRET_ACCESS_KEY"))
        return data

    def _run_operation(self, operation, *args, **kw):
        d = operation.run(*args, **kw)
        d.addErrback(convert_unknown_error)
        return d

    def connect(self, share=False):
        """Connect to the zookeeper ensemble running in the machine provider.

        @param share: Requests sharing of the connection with other clients
            attempting to connect to the same provider, if that's feasible.

        returns an open C{txzookeeper.client.ZookeeperClient} and a
        C{ensemble.storage.connection.TunnelProtocol}
        """
        connect = EC2Connect(self)
        return self._run_operation(connect, share=share)

    def get_file_storage(self):
        """Retrieve an S3-backed C{FileStorage}."""
        return FileStorage(self.s3, self.config["control-bucket"])

    def start_machine(self, machine_data, master=False):
        """Start a machine in the provider.

        @param machine_data: a dictionary of data to pass along to the newly
           launched machine.

        @param master: if True, machine will initialize the ensemble admin
            and run a provisioning agent.
        """
        return EC2LaunchMachine(self, master).run(machine_data)

    @inlineCallbacks
    def get_machines(self, instance_ids=()):
        """List machines running in the provider.

        @param instance_ids: ids of instances you want to get. Leave blank
            to list all machines owned by ensemble.

        @return: a list of EC2ProviderMachines.

        @raise: MachinesNotFound
        """
        group_name = "ensemble-%s" % self.environment_name
        try:
            instances = yield self.ec2.describe_instances(*instance_ids)
        except EC2Error as error:
            code = error.get_error_codes()
            if code == "InvalidInstanceID.NotFound":
                message = error.get_error_messages()
                raise MachinesNotFound(re.findall(
                        r"\bi-[0-9a-f]{3,15}\b", message))
            raise ProviderInteractionError(
                "EC2 error when looking up instances %s: %s"
                % (", ".join(instance_ids), error.get_error_messages()))

        machines = []
        for instance in instances:
            if instance.instance_state not in ("running", "pending"):
                continue
            if group_name not in instance.reservation.groups:
                continue
            machines.append(machine_from_instance(instance))

        if instance_ids:
            # We were asked for a specific list of machines, and if we can't
            # completely fulfil that request we should blow up.
            found_instance_ids = set(m.instance_id for m in machines)
            missing = set(instance_ids) - found_instance_ids
            if missing:
                raise MachinesNotFound(missing)
        returnValue(machines)

    @inlineCallbacks
    def shutdown_machines(self, requested_machines=()):
        """Terminate machine resources associated with this provider."""
        for machine in requested_machines:
            if not isinstance(machine, EC2ProviderMachine):
                raise ProviderError("Can only shut down EC2ProviderMachines; "
                                    "got a %r" % type(machine))

        requested_ids = [m.instance_id for m in requested_machines]
        killable_machines = yield self.get_machines(requested_ids)
        if not killable_machines:
            return  # Nothing to do

        killable_ids = [m.instance_id for m in killable_machines]
        terminated = yield self.ec2.terminate_instances(*killable_ids)

        # Pass on what was actually terminated, in the case the
        # machine has somehow disappeard since get_machines
        # above. This is to avoid getting EC2Error: Error Message:
        # Invalid id when running ec2.describe_instances in
        # remove_security_groups
        terminated_ids = [info[0] for info in terminated]
        yield remove_security_groups(self, terminated_ids)
        returnValue(killable_machines)

    def open_port(self, machine, machine_id, port, protocol="tcp"):
        """Authorizes `port` using `protocol` on EC2 for `machine`."""
        return open_provider_port(self, machine, machine_id, port, protocol)

    def close_port(self, machine, machine_id, port, protocol="tcp"):
        """Revokes `port` using `protocol` on EC2 for `machine`."""
        return close_provider_port(self, machine, machine_id, port, protocol)

    def get_opened_ports(self, machine, machine_id):
        """Returns a set of open (port, proto) pairs for `machine`."""
        return get_provider_opened_ports(self, machine, machine_id)
