"""
Ec2 Provider File Storage on S3
"""
from cStringIO import StringIO

from twisted.internet.defer import fail
from txaws.s3.exception import S3Error

from ensemble.errors import FileNotFound

_FILENOTFOUND_CODES = ("NoSuchKey", "NoSuchBucket")


def _safe_string(s):
    if isinstance(s, unicode):
        s = s.encode('utf8')
    return s


class FileStorage(object):

    def __init__(self, s3, bucket):
        self._s3 = s3
        self._bucket = bucket

    def get(self, name):
        """
        If the file exists in the storage, this method will download the file
        and return a file object. If it doesn't exist then a FileNotFound
        failure is returned.

        @return An open file object
        @type Deferred
        """
        # for now we do the simplest thing and just fetch the file
        # in a single call, s3 limits this to some mb, as we grow
        # to have formulas that might grow to this size, we can
        # revisit fetching in batches (head request for size) and
        # streaming to disk.
        d = self._s3.get_object(self._bucket, _safe_string(name))

        def on_content_retrieved(content):
            return StringIO(content)
        d.addCallback(on_content_retrieved)

        def on_no_file(failure):
            # Trap file not found errors and wrap them in an application error.
            failure.trap(S3Error)
            if failure.value.get_error_code() not in _FILENOTFOUND_CODES:
                return failure
            return fail(FileNotFound("s3://%s/%s" % (self._bucket, name)))
        d.addErrback(on_no_file)

        return d

    def put(self, remote_path, file_object):
        """
        Upload data to S3.

        @param remote_path: key on which to store the content

        @param file_object: open file object containing the content

        @return A deferred with a boolean value
        @type Deferred
        """
        content = file_object.read()
        path = _safe_string(remote_path)

        def put(ignored):
            return self._s3.put_object(self._bucket, path, content)

        def create_retry(failure):
            failure.trap(S3Error)
            if failure.value.get_error_code() != "NoSuchBucket":
                return failure
            d = self._s3.create_bucket(self._bucket)
            d.addCallback(put)
            return d

        d = put(None)
        d.addErrback(create_retry)
        return d
