from twisted.internet.defer import inlineCallbacks, returnValue
from txaws.ec2.exception import EC2Error

from ensemble.errors import ProviderInteractionError
from ensemble.providers.common.launch import LaunchMachine

from .machine import machine_from_instance
from .utils import log, get_launch_options


class EC2LaunchMachine(LaunchMachine):
    """Amazon EC2 operation for launching an instance"""

    @inlineCallbacks
    def start_machine(self, machine_variables, machine_id):
        """Actually launch an instance on EC2.

        `machine_variables`: non-provider-specific data, sufficient to
            define the machine's behaviour once it exists.

        `machine_id`: the external machine ID (also exists in the
            above bag of data)

        Returns a singe-entry list containing a ProviderMachine for the
        new instance
        """
        # Retrieves standard ec2 run_instances arguments, of note
        # image id, and instance type.
        machine_options = yield get_launch_options(
            self._provider.config,
            **machine_variables)

        # Ensure the ensemble security groups exist (both the overall
        # group and the machine-specific group) and are included in
        # the machine instance launch options.
        machine_options["security_groups"] = yield self._ensure_group(
            machine_options["security_groups"],
            machine_id)

        # Launch the instance.
        instances = yield self._provider.ec2.run_instances(**machine_options)
        returnValue([machine_from_instance(instance)
                     for instance in instances])

    def get_instance_id_command(self):
        """Snippet to discover local instance-id

        @return: a snippet of shell script that evaluates to the
            local machine's instance_id.
        """
        return "$(curl http://169.254.169.254/1.0/meta-data/instance-id)"

    def get_machine_variables(self, vars_in):
        vars_deferred = super(
            EC2LaunchMachine, self).get_machine_variables(vars_in)

        def set_provider_type(vars):
            vars["ensemble-provider-type"] = "ec2"
            return vars
        vars_deferred.addCallback(set_provider_type)
        return vars_deferred

    @inlineCallbacks
    def _ensure_group(self, groups, machine_id):
        """Ensure the ensemble group is the machine launch groups.

        Machines launched by ensemble are tagged with a group so they
        can be distinguished from other machines that might be running
        on an EC2 account. This group can be specified explicitly or
        implicitly defined by the environment name. In addition, a
        specific machine security group is created for each machine,
        so that its firewall rules can be configured per machine.

        `groups`: The configured security groups for a machine instance.

        `machine_id`: The external machine ID of this machine instance
        """
        # Label the instance with the ensemble name.
        security_groups = yield self._provider.ec2.describe_security_groups()
        ensemble_group = "ensemble-%s" % self._provider.environment_name
        ensemble_machine_group = "ensemble-%s-%s" % (
            self._provider.environment_name, machine_id)
        group_ids = [group.name for group in security_groups]

        # Ensure the provider group is added to the instance groups.
        if not ensemble_group in groups:
            groups.append(ensemble_group)

        # Create the provider group if doesn't exist.
        if not ensemble_group in group_ids:
            log.debug("Creating ensemble provider group %s", ensemble_group)
            yield self._provider.ec2.create_security_group(
                ensemble_group,
                "Ensemble group for %s" % self._provider.environment_name)

            # Authorize SSH.
            yield self._provider.ec2.authorize_security_group(
                ensemble_group,
                ip_protocol="tcp",
                from_port="22", to_port="22",
                cidr_ip="0.0.0.0/0")

            # We need to describe the group to pickup the owner_id for auth.
            groups_info = yield self._provider.ec2.describe_security_groups(
                ensemble_group)

            # Authorize Internal ZK Traffic
            yield self._provider.ec2.authorize_security_group(
                ensemble_group,
                source_group_name=ensemble_group,
                source_group_owner_id=groups_info.pop().owner_id)

        # Create the machine-specific group, but first see if there's
        # one already existing from a previous machine launch;
        # if so, delete it, since it can have the wrong firewall setup
        if ensemble_machine_group in group_ids:
            try:
                yield self._provider.ec2.delete_security_group(
                    ensemble_machine_group)
                log.debug("Deleted existing machine group %s, will replace",
                          ensemble_machine_group)
            except EC2Error, e:
                log.debug("Cannot delete security group %s: %s",
                          ensemble_machine_group, e)
                raise ProviderInteractionError(
                    "EC2 error when attempting to delete "
                    "security group %s: %s" % (
                    ensemble_machine_group, e))
        log.debug("Creating ensemble machine security group %s",
                  ensemble_machine_group)
        yield self._provider.ec2.create_security_group(
            ensemble_machine_group,
            "Ensemble group for %s machine %s" % (
                self._provider.environment_name, machine_id))
        groups.append(ensemble_machine_group)

        returnValue(groups)
