from yaml import load

from twisted.internet.defer import inlineCallbacks, succeed

from txaws.ec2.model import Instance, SecurityGroup

from ensemble.errors import EnvironmentNotFound, ProviderInteractionError
from ensemble.providers.common.launch import (
    DEFAULT_REPOSITORIES, DEFAULT_PACKAGES)
from ensemble.providers.ec2.machine import EC2ProviderMachine

from ensemble.lib.testing import TestCase
from ensemble.lib.mocker import MATCH

from .common import EC2TestMixin, EC2MachineLaunchMixin, MATCH_AMI


class EC2MachineLaunchTest(EC2TestMixin, EC2MachineLaunchMixin, TestCase):

    def _mock_launch(self, instance=None, custom_verify=None, machine_id=None):

        def verify_user_data(data):
            lines = data.split("\n")
            self.assertEqual(lines.pop(0), "#cloud-config")
            config = load("\n".join(lines))
            repos = [dict(source=r) for r in DEFAULT_REPOSITORIES]
            self.assertEqual(config["apt_sources"], repos)
            self.assertEqual(config["packages"], DEFAULT_PACKAGES)
            self.assertEqual(config["machine-data"]["machine-id"], "machine-1")

            if custom_verify:
                custom_verify(config)
            return True

        self.ec2.run_instances(
            image_id=MATCH_AMI,
            instance_type="m1.small",
            max_count=1,
            min_count=1,
            security_groups=[
                "%s-%s" % ("ensemble", self.env_name),
                "%s-%s-%s" % ("ensemble", self.env_name, machine_id)],
            user_data=MATCH(verify_user_data))

        if instance:
            self.mocker.result(succeed([instance]))
        else:
            self.mocker.result(succeed([]))

    def test_provider_launch(self):
        """
        The provider can be used to launch a machine with a minimal set of
        required packages, repositories, and and security groups.
        """
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_create_machine_group("machine-1")
        self._mock_launch_utils(region="us-east-1")
        self._mock_get_zookeeper_hosts()
        self._mock_launch(
            self.get_instance("i-foobar"), machine_id="machine-1")
        self.mocker.replay()

        def verify_result(result):
            (machine,) = result
            self.assert_machine(machine, "i-foobar", "")
        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        d.addCallback(verify_result)
        return d

    @inlineCallbacks
    def test_provider_launch_existing_security_group(self):
        """Verify that the launch works if the env security group exists"""
        instance = Instance("i-foobar", "running", dns_name="x1.example.com")
        security_group = SecurityGroup("ensemble-moon", "some description")

        self.ec2.describe_security_groups()
        self.mocker.result(succeed([security_group]))
        self._mock_create_machine_group("machine-1")
        self._mock_launch_utils(region="us-east-1")
        self._mock_get_zookeeper_hosts()
        self._mock_launch(instance, machine_id="machine-1")
        self.mocker.replay()

        provider = self.get_provider()
        provided_machines = yield provider.start_machine(
            {"machine-id": "machine-1"})
        self.assertEqual(len(provided_machines), 1)
        self.assertTrue(isinstance(provided_machines[0], EC2ProviderMachine))
        self.assertEqual(
            provided_machines[0].instance_id, instance.instance_id)

    @inlineCallbacks
    def test_provider_launch_existing_machine_security_group(self):
        """Verify that the launch works if the machine security group exists"""
        instance = Instance("i-foobar", "running", dns_name="x1.example.com")
        machine_group = SecurityGroup(
            "ensemble-moon-machine-1", "some description")

        self.ec2.describe_security_groups()
        self.mocker.result(succeed([machine_group]))
        self._mock_create_group()
        self._mock_delete_machine_group("machine-1")  # delete existing sg
        self._mock_create_machine_group("machine-1")  # then recreate
        self._mock_launch_utils(region="us-east-1")
        self._mock_get_zookeeper_hosts()
        self._mock_launch(instance, machine_id="machine-1")
        self.mocker.replay()

        provider = self.get_provider()
        provided_machines = yield provider.start_machine(
            {"machine-id": "machine-1"})
        self.assertEqual(len(provided_machines), 1)
        self.assertTrue(isinstance(provided_machines[0], EC2ProviderMachine))
        self.assertEqual(
            provided_machines[0].instance_id, instance.instance_id)

    @inlineCallbacks
    def test_provider_launch_existing_machine_security_group_is_active(self):
        """Verify launch fails properly if the machine group is stil active.

        This condition occurs when there is a corresponding machine in
        that security group, generally because it is still shutting
        down."""
        machine_group = SecurityGroup(
            "ensemble-moon-machine-1", "some description")

        self.ec2.describe_security_groups()
        self.mocker.result(succeed([machine_group]))
        self._mock_create_group()
        self._mock_delete_machine_group_was_deleted("machine-1")  # sg is gone!
        self._mock_launch_utils(region="us-east-1")
        self._mock_get_zookeeper_hosts()
        self.mocker.replay()

        provider = self.get_provider()
        ex = yield self.assertFailure(
            provider.start_machine({"machine-id": "machine-1"}),
            ProviderInteractionError)
        self.assertEqual(
            str(ex),
            self.get_wrapped_ec2_error_text(
                "ensemble-moon-machine-1", "delete security group",
                "There are active instances using security group %r"
                ))

    def test_provider_type_machine_variable(self):
        """The provider type is available via cloud-init."""
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_create_machine_group("machine-1")
        self._mock_launch_utils(region="us-east-1")
        self._mock_get_zookeeper_hosts()

        def verify_provider_type(data):
            self.assertEqual(
                data["machine-data"]["ensemble-provider-type"], "ec2")

        self._mock_launch(custom_verify=verify_provider_type,
                          machine_id="machine-1")
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        return d

    def test_launch_with_no_ensemble_s3_state(self):
        # XXX not really testing anything EC2-specific
        """
        Attempting to launch without any ensemble saved state, means
        we can't provide a way for a launched instance to connect
        to the zookeeper shared state. An assertion error is raised
        instead of allowing this.
        """
        self._mock_get_zookeeper_hosts(False)
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        self.assertFailure(d, EnvironmentNotFound)
        return d

    def test_launch_with_no_ensemble_zookeeper_hosts(self):
        # XXX not really testing anything EC2-specific
        """
        Attempting to launch without any ensemble zookeeper hosts, means
        we can't provide a way for a launched instance to connect
        to the zookeeper shared state. An assertion error is raised
        instead of allowing this.
        """
        self._mock_get_zookeeper_hosts([])
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        self.assertFailure(d, EnvironmentNotFound)
        return d

    def test_launch_starts_machine_agent(self):
        # XXX not really testing anything EC2-specific
        """A launch'd machine should have a machine agent."""
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_create_machine_group("machine-1")
        self._mock_launch_utils(region="us-east-1")
        self._mock_get_zookeeper_hosts()

        def verify_machine_agent(data):
            script = ("ENSEMBLE_MACHINE_ID=machine-1 "
                      "ENSEMBLE_ZOOKEEPER=es.example.internal:2181 "
                      "python -m ensemble.agents.machine "
                      "-n --logfile=/var/log/ensemble/machine-agent.log "
                      "--pidfile=/var/run/ensemble/machine-agent.pid")
            self.assertEqual(script, data["runcmd"][-1])

        self._mock_launch(custom_verify=verify_machine_agent,
                          machine_id="machine-1")
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        return d

    def _mock_launch_with_ami_params(self, get_ami_kwargs):
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_create_machine_group("machine-1")
        self._mock_launch_utils(**get_ami_kwargs)
        self._mock_get_zookeeper_hosts()
        self._mock_launch(
            self.get_instance("i-foobar"), machine_id="machine-1")

    def test_launch_options_known_ami(self):
        self._mock_launch_with_ami_params({})
        self.mocker.replay()

        provider = self.get_provider()
        # TODO shouldn't something be testing we actually launch
        # an instance of the expected AMI?
        provider.config["default-image-id"] = "ami-XXXXXX"
        return provider.start_machine({"machine-id": "machine-1"})

    def test_launch_options_region(self):
        self._mock_launch_with_ami_params({"region": "somewhere-else-1"})
        self.mocker.replay()

        provider = self.get_provider()
        provider.config["region"] = "somewhere-else-1"
        return provider.start_machine({"machine-id": "machine-1"})

    def test_launch_options_other(self):
        self._mock_launch_with_ami_params({
            "region": "us-east-1",
            "ubuntu_release_name": "choleric",
            "arch": "quantum86",
            "ebs": "maybe",
            "daily": "perhaps"})
        self.mocker.replay()

        provider = self.get_provider()
        return provider.start_machine({
            "machine-id": "machine-1",
            "image_release_name": "choleric",
            "image_arch": "quantum86",
            "image_ebs": "maybe",
            "image_daily": "perhaps"})
