from base64 import b64encode
from cStringIO import StringIO
from gzip import GzipFile

from twisted.internet.defer import inlineCallbacks, returnValue

from ensemble.providers.common.launch import LaunchMachine
from ensemble.providers.common.utils import format_cloud_init
from ensemble.providers.orchestra.machine import machine_from_dict

_INSTANCE_ID_VAR = "ENSEMBLE_INSTANCE_ID"

_SET_INSTANCE_ID_TEMPLATE = "export %s=%%s" % _INSTANCE_ID_VAR

_CLOUD_INIT_KEYS = "authorized_keys packages repositories scripts".split()

_LATE_COMMAND_TEMPLATE = """
seed_d=/var/lib/cloud/seed/nocloud-net
mkdir -p "$seed_d"
cat > "$seed_d"/meta-data <<"EOF"
instance-id: %s
EOF

cat > "$seed_d"/user-data <<"EOF"
%s
EOF
"""

_KSMETA_LATE_COMMAND_TEMPLATE = (
    "in-target sh -c '"
        "f=$1; "
        "shift; "
        "echo $0 | base64 --decode | gunzip > $f "
        "&& chmod u+x $f "
        "&& $f $*"
    "' %s /root/late-command")


def _cloud_init_vars(raw_vars):
    # XXX this is nasty; see bug#820892
    vars = raw_vars.copy()
    data = {}
    for key in vars.keys():
        if key not in _CLOUD_INIT_KEYS:
            data[key] = vars.pop(key)
    vars["data"] = data
    return vars


def _base64_gzip(content):
    gzipped = StringIO()
    gzip_file = GzipFile(fileobj=gzipped, mode="wb", compresslevel=9)
    gzip_file.write(content)
    gzip_file.close()
    return b64encode(gzipped.getvalue())


def _ks_meta(**kwargs):
    return " ".join('%s="%s"' % item for item in kwargs.items())


class OrchestraLaunchMachine(LaunchMachine):

    @inlineCallbacks
    def start_machine(self, variables, machine_id):
        cobbler = self._provider.cobbler
        instance_id = yield cobbler.acquire_system()
        ks_meta = self._build_ks_meta(instance_id, variables)
        yield cobbler.set_on_system(instance_id, "ks_meta", ks_meta)
        yield cobbler.start_system(instance_id)
        info = yield cobbler.describe_system(instance_id)
        returnValue([machine_from_dict(info)])

    def get_instance_id_command(self):
        return "$" + _INSTANCE_ID_VAR

    def _build_ks_meta(self, instance_id, variables):
        # XXX it's a little bit dirty to modify scripts at this stage;
        # but we don't actually know the instance_id at get_variables time.
        set_instance_id = _SET_INSTANCE_ID_TEMPLATE % instance_id
        variables["scripts"].insert(0, set_instance_id)

        cloud_init_vars = _cloud_init_vars(variables)
        user_data = format_cloud_init(**cloud_init_vars)
        late_command = _LATE_COMMAND_TEMPLATE % (instance_id, user_data)
        encoded = _base64_gzip(late_command)
        ensemble_late_command = _KSMETA_LATE_COMMAND_TEMPLATE % encoded
        return _ks_meta(MACHINE_ID=variables["machine-id"],
                        ENSEMBLE_LATE_COMMAND=ensemble_late_command)
