from cStringIO import StringIO

from twisted.internet.defer import fail, succeed
from twisted.web.error import Error

from ensemble.errors import FileNotFound
from ensemble.lib.testing import TestCase
from ensemble.providers.orchestra import MachineProvider


def get_file_storage(with_storage_url=True):
    config = {"storage-url": "http://somewhe.re/webdav",
              "orchestra-server": "somewhereel.se",
              "orchestra-user": "user",
              "orchestra-pass": "pass",
              "acquired-mgmt-class": "acquired",
              "available-mgmt-class": "available"}
    if not with_storage_url:
        del config["storage-url"]
    provider = MachineProvider("blah", config)
    return provider.get_file_storage()


class FileStorageTest(TestCase):

    def test_get_works_no_storage_url(self):
        getPage = self.mocker.replace("twisted.web.client.getPage")
        getPage("http://somewhereel.se/webdav/rubber/chicken")
        self.mocker.result(succeed("pulley"))
        self.mocker.replay()

        fs = get_file_storage(with_storage_url=False)
        d = fs.get("rubber/chicken")

        def verify(result):
            self.assertEquals(result.read(), "pulley")
        d.addCallback(verify)
        return d

    def test_get_works(self):
        getPage = self.mocker.replace("twisted.web.client.getPage")
        getPage("http://somewhe.re/webdav/rubber/chicken")
        self.mocker.result(succeed("pulley"))
        self.mocker.replay()

        fs = get_file_storage()
        d = fs.get("rubber/chicken")

        def verify(result):
            self.assertEquals(result.read(), "pulley")
        d.addCallback(verify)
        return d

    def test_get_fails(self):
        getPage = self.mocker.replace("twisted.web.client.getPage")
        getPage("http://somewhe.re/webdav/rubber/chicken")
        self.mocker.result(fail(Error("404")))
        self.mocker.replay()

        fs = get_file_storage()
        d = fs.get("rubber/chicken")
        self.assertFailure(d, FileNotFound)
        return d

    def test_get_errors(self):
        getPage = self.mocker.replace("twisted.web.client.getPage")
        getPage("http://somewhe.re/webdav/rubber/chicken")
        self.mocker.result(fail(Error("500")))
        self.mocker.replay()

        fs = get_file_storage()
        d = fs.get("rubber/chicken")
        self.assertFailure(d, Error)
        return d

    def test_put_works(self):
        getPage = self.mocker.replace("twisted.web.client.getPage")
        getPage("http://somewhe.re/webdav/rubber/chicken",
                method="PUT", postdata="pulley")
        self.mocker.result(succeed(None))
        self.mocker.replay()

        fs = get_file_storage()
        d = fs.put("rubber/chicken", StringIO("pulley"))

        def verify(result):
            self.assertEquals(result, True)
        d.addCallback(verify)
        return d

    def test_put_works_no_storage_url(self):
        getPage = self.mocker.replace("twisted.web.client.getPage")
        getPage("http://somewhereel.se/webdav/rubber/chicken",
                method="PUT", postdata="pulley")
        self.mocker.result(succeed(None))
        self.mocker.replay()

        fs = get_file_storage(with_storage_url=False)
        d = fs.put("rubber/chicken", StringIO("pulley"))

        def verify(result):
            self.assertEquals(result, True)
        d.addCallback(verify)
        return d

    def test_put_handles_204(self):
        """If we're overwriting instead of creating, we get 204 instead of 200
        """
        getPage = self.mocker.replace("twisted.web.client.getPage")
        getPage("http://somewhe.re/webdav/rubber/chicken",
                method="PUT", postdata="pulley")
        self.mocker.result(fail(Error("204")))
        self.mocker.replay()

        fs = get_file_storage()
        d = fs.put("rubber/chicken", StringIO("pulley"))

        def verify(result):
            self.assertEquals(result, True)
        d.addCallback(verify)
        return d

    def test_put_errors(self):
        getPage = self.mocker.replace("twisted.web.client.getPage")
        getPage("http://somewhe.re/webdav/rubber/chicken",
                method="PUT", postdata="pulley")
        self.mocker.result(fail(Error("500")))
        self.mocker.replay()

        fs = get_file_storage()
        d = fs.put("rubber/chicken", StringIO("pulley"))
        self.assertFailure(d, Error)
        return d
