from ensemble.errors import (
    EnsembleError, FileNotFound, FileAlreadyExists, InvalidEnsembleHeaderValue,
    NoConnection, InvalidHost, InvalidUser, ProviderError,
    ProviderInteractionError, CannotTerminateMachine, MachinesNotFound,
    EnvironmentPending, EnvironmentNotFound, IncompatibleVersion)

from ensemble.lib.testing import TestCase


class ErrorsTest(TestCase):

    def assertIsEnsembleError(self, error):
        self.assertTrue(isinstance(error, EnsembleError),
                        "%s is not a subclass of EnsembleError" %
                        error.__class__.__name__)

    def test_IncompatibleVersion(self):
        error = IncompatibleVersion(123, 42)
        self.assertEqual(
            str(error),
            "Incompatible ensemble protocol versions (found 123, want 42)")
        self.assertIsEnsembleError(error)

    def test_FileNotFound(self):
        error = FileNotFound("/path")
        self.assertEquals(str(error), "File was not found: /path")
        self.assertIsEnsembleError(error)

    def test_FileAlreadyExists(self):
        error = FileAlreadyExists("/path")
        self.assertEquals(str(error),
                          "File already exists, won't overwrite: /path")
        self.assertIsEnsembleError(error)

    def test_InvalidEnsembleHeaderValue(self):
        error = InvalidEnsembleHeaderValue("/path", "xpctd", "fnd")
        self.assertEquals(str(error),
                          "Expected a YAML file with an 'ensemble: xpctd' "
                          "header, found 'ensemble: fnd': /path")
        self.assertIsEnsembleError(error)

    def test_NoConnection(self):
        error = NoConnection("unable to connect")
        self.assertIsEnsembleError(error)

    def test_InvalidHost(self):
        error = InvalidHost("Invalid host for SSH forwarding")
        self.assertTrue(isinstance(error, NoConnection))
        self.assertEquals(
            str(error),
            "Invalid host for SSH forwarding")

    def test_InvalidUser(self):
        error = InvalidUser("Invalid SSH key")
        self.assertTrue(isinstance(error, NoConnection))
        self.assertEquals(
            str(error),
            "Invalid SSH key")

    def test_ProviderError(self):
        error = ProviderError("Invalid credentials")
        self.assertIsEnsembleError(error)

    def test_ProviderInteractionError(self):
        error = ProviderInteractionError(OSError("Bad Stuff"))
        self.assertIsEnsembleError(error)
        self.assertEquals(
            str(error),
            "ProviderError: Interaction with machine provider failed: "
            "OSError('Bad Stuff',)")

    def test_CannotTerminateMachine(self):
        error = CannotTerminateMachine(0, "environment would be destroyed")
        self.assertIsEnsembleError(error)
        self.assertEquals(
            str(error),
            "Cannot terminate machine 0: environment would be destroyed")

    def test_MachinesNotFoundSingular(self):
        error = MachinesNotFound(("i-sublimed",))
        self.assertEquals(error.instance_ids, ["i-sublimed"])
        self.assertEquals(str(error),
                          "Cannot find machine: i-sublimed")

    def test_MachinesNotFoundPlural(self):
        error = MachinesNotFound(("i-disappeared", "i-exploded"))
        self.assertEquals(error.instance_ids, ["i-disappeared", "i-exploded"])
        self.assertEquals(str(error),
                          "Cannot find machines: i-disappeared, i-exploded")

    def testEnvironmentNotFoundWithInfo(self):
        error = EnvironmentNotFound("problem")
        self.assertEquals(str(error),
                          "Ensemble environment not found: problem")

    def testEnvironmentNotFoundNoInfo(self):
        error = EnvironmentNotFound()
        self.assertEquals(str(error),
                          "Ensemble environment not found: no details "
                          "available")

    def testEnvironmentPendingWithInfo(self):
        error = EnvironmentPending("problem")
        self.assertEquals(str(error),
                          "Ensemble environment is not accessible: problem")

    def testEnvironmentPendingNoInfo(self):
        error = EnvironmentPending()
        self.assertEquals(str(error),
                          "Ensemble environment is not accessible: no "
                          "details available")
